# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

cumulative_prop <- function(input, top) {
    .Call(`_scuttle_cumulative_prop`, input, top)
}

downsample_column <- function(input, prop) {
    .Call(`_scuttle_downsample_column`, input, prop)
}

downsample_matrix <- function(rmat, total, required) {
    .Call(`_scuttle_downsample_matrix`, rmat, total, required)
}

fit_linear_model <- function(qr, qraux, exprs, get_coefs) {
    .Call(`_scuttle_fit_linear_model`, qr, qraux, exprs, get_coefs)
}

sum_row_counts <- function(counts, genes, runs) {
    .Call(`_scuttle_sum_row_counts`, counts, genes, runs)
}

